bl_info = {
    "name": "Better Pivot Tools",
    "author": "Danel Sánchez",
    "version": (1, 0),
    "blender": (3, 0, 0),
    "location": "View3D > Edit/Object Mode",
    "description": "Tool for fast pivot manipulation.",
    "category": "Mesh"
}

import bpy
import bmesh
from bpy.types import Operator, Panel
from bpy.props import EnumProperty, StringProperty
import mathutils

def restore_cursor(context, cursor_loc, cursor_rot):
    context.scene.cursor.location = cursor_loc
    context.scene.cursor.rotation_euler = cursor_rot

def get_lowest_vertex_world_position(obj):
    bm = bmesh.new()
    bm.from_mesh(obj.data)

    lowest_vertex = min(
        [obj.matrix_world @ v.co for v in bm.verts], 
        key=lambda x: x.z
    )

    bm.free()

    return lowest_vertex

def get_object_center(obj, method='world'):
    if method == 'world':
        return obj.matrix_world.translation

    elif method == 'origin':
        return obj.location

    elif method == 'bounds':
        bbox_corners = [obj.matrix_world @ mathutils.Vector(corner) for corner in obj.bound_box]
        return sum(bbox_corners, mathutils.Vector()) / len(bbox_corners)

    elif method == 'geometry':
        bm = bmesh.new()
        bm.from_mesh(obj.data)

        center = sum((obj.matrix_world @ v.co for v in bm.verts), mathutils.Vector()) / len(bm.verts)

        bm.free()
        return center

def transform_to_delta(obj):
    pos = obj.location.copy()
    rot = obj.rotation_euler.copy()
    obj.rotation_euler = (0, 0, 0)
    obj.location = (0, 0, 0)
    obj.delta_rotation_euler = rot
    obj.delta_location = pos

class PIVOT_TOOL_OT_MovePivotOperator(Operator):
    """Operator to move the pivot of an object based on the selected mode."""
    bl_idname = "pivot_tool.basic_pivot_ops"
    bl_label = "Basic Pivot Tools"
    bl_options = {'UNDO'}

    mode: EnumProperty(
        items=[
            ('COMPONENT', "Gizmo", "Move pivot to gizmo position"),
            ('BOTTOM', "Bottom", "Move pivot to lowest Z position"),
            ('CENTER', "Center Z", "Reset pivot to object center"),
        ],
        name="Mode",
        default='COMPONENT'
    )

    @classmethod
    def poll(cls, context):
        obj = context.active_object
        return obj and obj.select_get()

    def execute(self, context):
        obj = context.active_object
        if not obj:
            self.report({'ERROR'}, "No active object found.")
            return {'CANCELLED'}

        cursor_loc = context.scene.cursor.location.copy()
        cursor_rot = context.scene.cursor.rotation_euler.copy()

        if self.mode == 'COMPONENT':
            if obj.mode != 'EDIT':
                self.report({'ERROR'}, "Gizmo mode requires Edit Mode. Please switch to Edit Mode.")
                return {'CANCELLED'}

            bm = bmesh.from_edit_mesh(obj.data)
            selected_verts = [v for v in bm.verts if v.select]

            if not selected_verts:
                self.report({'ERROR'}, "No vertices selected in Edit Mode.")
                return {'CANCELLED'}

            median_local = sum((v.co for v in selected_verts), mathutils.Vector()) / len(selected_verts)
            median_world = obj.matrix_world @ median_local
            context.scene.cursor.location = median_world
            bpy.ops.wm.tool_set_by_id(name='builtin.move', space_type='VIEW_3D')

        elif self.mode == 'BOTTOM':
            if not obj.data.vertices:
                self.report({'ERROR'}, "Object has no vertices.")
                return {'CANCELLED'}
            lowest_z = get_lowest_vertex_world_position(obj)
            new_position = mathutils.Vector((obj.matrix_world.translation.x, obj.matrix_world.translation.y, lowest_z.z))
            context.scene.cursor.location = new_position
            bpy.ops.wm.tool_set_by_id(name='builtin.move', space_type='VIEW_3D')

        elif self.mode == 'CENTER':
            obj = context.active_object
            obj.select_set(True)
            center = get_object_center(obj, 'geometry')
            context.scene.cursor.location = center
            bpy.ops.wm.tool_set_by_id(name='builtin.move', space_type='VIEW_3D')

        if self.mode in {'BOTTOM', 'CENTER'}:
            context.scene.cursor.rotation_euler = (0, 0, 0)
        else:
            context.scene.cursor.rotation_euler = obj.rotation_euler

        orig_mode = obj.mode
        if orig_mode == 'EDIT':
            bpy.ops.object.mode_set(mode='OBJECT')

        bpy.ops.object.origin_set(type='ORIGIN_CURSOR')

        if orig_mode == 'EDIT':
            bpy.ops.object.mode_set(mode='EDIT')

        restore_cursor(context, cursor_loc, cursor_rot)

        self.report({'INFO'}, f"Pivot moved to {self.mode} position.")
        return {'FINISHED'}

class PIVOT_TOOL_OT_CreateTempEmptyOperator(Operator):
    """Operator to create a temporary empty at the selected vertices' median position."""
    bl_idname = "pivot_tool.create_temp_empty"
    bl_label = "Create Temporary Empty"
    bl_options = {'UNDO'}

    @classmethod
    def poll(cls, context):
        obj = context.active_object
        return obj and obj.select_get()

    def execute(self, context):
        obj = context.active_object

        context.scene.pivot_tool_selected_obj = obj.name

        bpy.ops.object.mode_set(mode='OBJECT')
        bpy.ops.object.empty_add(type='PLAIN_AXES', location=bpy.context.active_object.location)
        temp_empty = context.active_object
        temp_empty.name = "TempEmpty"
        bpy.ops.wm.tool_set_by_id(name='builtin.move', space_type='VIEW_3D')

        self.report({'INFO'}, "Temporary empty created and selected.")
        return {'FINISHED'}

class PIVOT_TOOL_OT_CancelTempEmptyOperator(Operator):
    """Operator to cancel and delete the temporary empty."""
    bl_idname = "pivot_tool.cancel_temp_empty"
    bl_label = "Cancel Temporary Empty"
    bl_options = {'UNDO'}

    def execute(self, context):
        temp_empty = bpy.data.objects.get("TempEmpty")
        if temp_empty:
            bpy.data.objects.remove(temp_empty)
            selected_obj_name = context.scene.pivot_tool_selected_obj
            selected_obj = bpy.data.objects.get(selected_obj_name)
            bpy.ops.object.select_all(action='DESELECT')
            selected_obj.select_set(True)
            bpy.context.view_layer.objects.active = selected_obj
            self.report({'INFO'}, "Temporary empty deleted.")
        return {'FINISHED'}

class PIVOT_TOOL_OT_AcceptTempEmptyOperator(Operator):
    bl_idname = "pivot_tool.accept_temp_empty"
    bl_label = "Accept Temporary Empty"
    bl_options = {'UNDO'}

    def execute(self, context):
        temp_empty = bpy.data.objects.get("TempEmpty")
        if not temp_empty:
            self.report({'ERROR'}, "No temporary empty found.")
            return {'CANCELLED'}

        selected_obj_name = context.scene.pivot_tool_selected_obj
        selected_obj = bpy.data.objects.get(selected_obj_name)
        if not selected_obj:
            self.report({'ERROR'}, "Selected object not found.")
            return {'CANCELLED'}

        cursor = bpy.context.scene.cursor
        cursor_loc = context.scene.cursor.location.copy()
        cursor_rot = context.scene.cursor.rotation_euler.copy()

        cursor.location = temp_empty.location
        cursor.rotation_euler = temp_empty.rotation_euler

        selected_obj.name = selected_obj_name + "_old"
        bpy.ops.object.select_all(action='DESELECT')
        selected_obj.select_set(True)
        bpy.context.view_layer.objects.active = selected_obj
        bpy.ops.object.duplicate()

        duplicated_obj = bpy.context.active_object
        duplicated_obj.name = selected_obj_name
        duplicated_obj.select_set(True)
        bpy.context.view_layer.objects.active = duplicated_obj
        bpy.ops.object.origin_set(type='ORIGIN_CURSOR')
        duplicated_obj.rotation_euler = context.scene.cursor.rotation_euler

        bpy.ops.object.mode_set(mode='EDIT')
        bpy.ops.mesh.select_all(action='SELECT')
        bpy.ops.mesh.delete(type='VERT')
        bpy.ops.object.mode_set(mode='OBJECT')
        transform_to_delta(duplicated_obj)

        bpy.ops.object.select_all(action='DESELECT')
        duplicated_obj.select_set(True)
        selected_obj.select_set(True)
        bpy.context.view_layer.objects.active = duplicated_obj

        bpy.ops.object.join()
        bpy.context.view_layer.objects.active = duplicated_obj
        selected_obj = bpy.context.active_object

        cursor.location = cursor_loc
        cursor.rotation_euler = cursor_rot

        bpy.data.objects.remove(temp_empty)

        self.report({'INFO'}, "Pivot updated to temporary empty position and rotation.")
        return {'FINISHED'}

class PIVOT_TOOL_OT_ResetOrientationOperator(Operator):
    """Operator to reset the pivot's orientation to global (0, 0, 0) coordinates."""
    bl_idname = "pivot_tool.reset_orientation"
    bl_label = "Reset Orientation"
    bl_options = {'UNDO'}

    @classmethod
    def poll(cls, context):
        obj = context.active_object
        return obj and obj.select_get()

    def execute(self, context):
        obj = context.active_object
        obj.select_set(True)
        bpy.ops.object.transform_apply(location=False, rotation=True, scale=False)

        self.report({'INFO'}, "Pivot orientation reset to global (0, 0, 0).")
        return {'FINISHED'}

class PIVOT_TOOL_PT_Panel(Panel):
    """UI panel for the pivot manipulation tool."""
    bl_label = "Pivot Manipulation"
    bl_idname = "PIVOT_PT_Panel"
    bl_space_type = 'VIEW_3D'
    bl_region_type = 'UI'
    bl_category = "Tool"

    def draw(self, context):
        layout = self.layout
        scene = context.scene
        obj = context.active_object

        if not obj:
            layout.label(text="No active object", icon='ERROR')
            return

        row = layout.row()

        box = row.box()
        box.label(text="Object Mode", icon='FACESEL')
        temp_empty = bpy.data.objects.get("TempEmpty")
        if obj and obj.select_get():
            row.enabled = not temp_empty
        col = box.column(align=True)
        col.operator("pivot_tool.basic_pivot_ops", 
                    text="Bottom Pivot", 
                    icon='ANCHOR_BOTTOM').mode = 'BOTTOM'
        col.operator("pivot_tool.basic_pivot_ops", 
                    text="Center Pivot", 
                    icon='ANCHOR_CENTER').mode = 'CENTER'

        box = row.box()
        box.label(text="Edit Mode", icon='VERTEXSEL')

        temp_empty = bpy.data.objects.get("TempEmpty")
        row.enabled = not temp_empty
        col = box.column(align=True)
        col.operator("pivot_tool.basic_pivot_ops", 
                        text="Selection Pivot", 
                        icon='SELECT_SET').mode = 'COMPONENT'

        custom_box = layout.box()
        custom_box.label(text="Advanced Pivot Tools", icon='EMPTY_AXIS')
        row = custom_box.row()

        temp_empty = bpy.data.objects.get("TempEmpty")
        if obj and obj.select_get():
            row.enabled = not temp_empty
        row.operator("pivot_tool.create_temp_empty", 
                    text="Create Custom Pivot", 
                    icon='EMPTY_DATA')

        row.operator("pivot_tool.reset_orientation", 
                        text="Reset Pivot Orientation", 
                        icon='REMOVE')

        if temp_empty:
            row = custom_box.row()
            row.alert = True
            row.operator("pivot_tool.cancel_temp_empty", 
                        text="Cancel", 
                        icon='CANCEL')
            row.alert = False
            row.operator("pivot_tool.accept_temp_empty", 
                        text="Apply Pivot", 
                        icon='CHECKMARK')

addon_keymaps = []

def register():
    bpy.utils.register_class(PIVOT_TOOL_OT_MovePivotOperator)
    bpy.utils.register_class(PIVOT_TOOL_OT_CreateTempEmptyOperator)
    bpy.utils.register_class(PIVOT_TOOL_OT_CancelTempEmptyOperator)
    bpy.utils.register_class(PIVOT_TOOL_OT_AcceptTempEmptyOperator)
    bpy.utils.register_class(PIVOT_TOOL_OT_ResetOrientationOperator)
    bpy.utils.register_class(PIVOT_TOOL_PT_Panel)

    bpy.types.Scene.pivot_tool_selected_obj = StringProperty(
        name="Selected Object",
        description="Stores the name of the selected object before creating the temporary empty",
        default=""
    )

def unregister():
    for km, kmi in addon_keymaps:
        km.keymap_items.remove(kmi)
    addon_keymaps.clear()
    bpy.utils.unregister_class(PIVOT_TOOL_OT_MovePivotOperator)
    bpy.utils.unregister_class(PIVOT_TOOL_OT_CreateTempEmptyOperator)
    bpy.utils.unregister_class(PIVOT_TOOL_OT_CancelTempEmptyOperator)
    bpy.utils.unregister_class(PIVOT_TOOL_OT_AcceptTempEmptyOperator)
    bpy.utils.unregister_class(PIVOT_TOOL_OT_ResetOrientationOperator)
    bpy.utils.unregister_class(PIVOT_TOOL_PT_Panel)

    del bpy.types.Scene.pivot_tool_selected_obj
    for km, kmi in addon_keymaps:
        km.keymap_items.remove(kmi)
    addon_keymaps.clear()

if __name__ == "__main__":
    register()
